use crate::{middleware::AuthenticatedAdminMiddleware, AppState};
use actix_multipart::Multipart;
use actix_web::{post, web::Data, HttpRequest, HttpResponse, Responder};
use futures::StreamExt;
use std::{
    fs::{create_dir_all, remove_file, File},
    io::Write,
};

// TODO separate file writing logic from Multipart implementation and implement the unit tests
#[post("/post-files")]
pub async fn post_files(
    app_state: Data<AppState>,
    mut payload: Multipart,
    middleware: Data<AuthenticatedAdminMiddleware<'_>>,
    req: HttpRequest,
) -> impl Responder {
    if middleware.exec(&app_state, &req, None).await.is_err() {
        return HttpResponse::Unauthorized().finish();
    }

    let uploads_dir = app_state.env.public_dir.join("uploads");

    if !uploads_dir.exists() {
        create_dir_all(&uploads_dir).unwrap();
    }

    let mut uploaded_filepathes = Vec::new();

    while let Some(item) = payload.next().await {
        match item {
            Ok(mut field) => {
                // Field in turn is stream of *Bytes* object
                // A multipart/form-data stream has to contain `content_disposition`
                let content_disposition = field
                    .content_disposition()
                    .expect("Missing Content Disposition header");

                let filename = content_disposition
                    .get_filename()
                    .expect("Missing filename");

                let filepath = uploads_dir.join(&filename);

                let f = File::create(&filepath);

                if f.is_err() {
                    return HttpResponse::InternalServerError()
                        .body(format!("Error creating file {:?}", f));
                }

                let mut f = f.unwrap();
                let mut error = None;

                // Field in turn is stream of *Bytes* object
                'chunks: while let Some(chunk) = field.next().await {
                    match chunk {
                        Ok(chunk) => {
                            if f.write_all(&chunk).is_err() {
                                error = Some("Error writing chunk".to_string());
                                break 'chunks;
                            }
                        }
                        Err(e) => {
                            error = Some(format!("Error writing file {} : {:?}", filename, e));
                            break 'chunks;
                        }
                    }
                }

                if let Some(err) = error {
                    remove_file(&filepath).unwrap();
                    return HttpResponse::InternalServerError().body(err);
                }

                uploaded_filepathes.push(String::from(filepath.to_string_lossy()));
            }
            Err(e) => {
                return HttpResponse::InternalServerError().body(format!("FIELD ERR {:?}", e))
            }
        }
    }

    HttpResponse::Ok().json(uploaded_filepathes)
}

// EXAMPLE FROM ACTIX REPO

// use futures::TryStreamExt;
// use std::io::Write;

// pub async fn save_file(mut payload: Multipart) -> impl Responder {
//     println!("SAVE FILE");
//     // iterate over multipart stream
//     while let Some(mut field) = payload.try_next().await.unwrap() {
//         // A multipart/form-data stream has to contain `content_disposition`
//         let content_disposition = field.content_disposition().unwrap();
//         let filename = content_disposition.get_filename().unwrap();
//         let filepath = format!("./tmp/{filename}");

//         // File::create is blocking operation, use threadpool
//         let mut f = actix_web::web::block(|| std::fs::File::create(filepath))
//             .await
//             .unwrap();

//         // Field in turn is stream of *Bytes* object
//         while let Some(chunk) = field.try_next().await.unwrap() {
//             // filesystem operations are blocking, we have to use threadpool
//             f = actix_web::web::block(move || f.write_all(&chunk).map(|_| f))
//                 .await
//                 .unwrap();
//         }
//     }

//     HttpResponse::Ok().body("sucess")
// }

/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*@@
 *@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*@@
 *  _______   ______    ______   _______   *@@
 * |__   __@ |  ____@  /  ____@ |__   __@  *@@
 *    |  @   |  @__    \_ @_       |  @    *@@
 *    |  @   |   __@     \  @_     |  @    *@@
 *    |  @   |  @___   ____\  @    |  @    *@@
 *    |__@   |______@  \______@    |__@    *@@
 *                                         *@@
 *@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*@@
 *@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*@*/

#[cfg(test)]
mod test_static_files {
    use super::*;
    use crate::middleware::get_auth_cookie;
    use actix_web::{
        http::{Method, StatusCode},
        test, App,
    };

    #[tokio::test]
    async fn post_files_unauthenticated_should_be_unauthorized() {
        let app_state = AppState::for_test().await;

        let mut app = test::init_service(
            App::new()
                .app_data(Data::new(app_state.clone()))
                .app_data(Data::new(AuthenticatedAdminMiddleware::new(
                    "kuadrado-admin-auth",
                )))
                .service(post_files),
        )
        .await;

        let req = test::TestRequest::with_uri("/post-files")
            .method(Method::POST)
            .cookie(get_auth_cookie(
                "wrong-cookie",
                app_state.encryption.random_ascii_lc_string(32),
            ))
            .to_request();

        let resp = test::call_service(&mut app, req).await;
        assert_eq!(resp.status(), StatusCode::UNAUTHORIZED);
    }
}
